/*
	IndependentEQ.c: Routines that deal with linearly independent
					 reaction equations.

	RedMech is a tool for automatically reducing CnHm-mechanisms.
	Written by J. Goettgens, P. Terhoeven, R. Korff	
	
	The SVD code was taken from the book "Numerical Recipes in C",
	W.H. Press, B.P. Flannery, S.A. Teukolsky, and W.T. Vetterling,
	Cambridge University Press (1988)
*/

#ifdef applec
#pragma load "RedMech.dump"
#else
#include "RedMechHeaders.h"
#endif

#include "svd.h"

int IndependentEQ( MatrixPtr m )
{
	int			i, j, rows = m->rows, cols = m->cols, newRows;
	int			indepEQ = 0;
	double		**a = m->mat;
	double		**u = NULL, **v = NULL, *w = NULL, *vr1 = NULL;
	
	/*	Make sure we have at least as much rows as columns.						*/
	if ( cols > rows ) newRows = cols;
	else newRows = rows;

	u = New2DArray( newRows, cols );
	v = New2DArray( newRows, cols );
	w = New1DArray( newRows );

	for ( i = 0; i < rows; ++i ) {
		for ( j = 0; j < cols; ++j ) {
			u[i][j] = a[i][j];
		}
	}

	svdcmp( u, newRows, cols, w, v, NULL );
	indepEQ = svrank( w, newRows, kTiny );
	
	/*fprintf( stderr, "rank is %d\n", indepEQ );*/
	
	Free1DArray( w );
	Free2DArray( v );
	Free2DArray( u );
	
	return indepEQ;
}


MatrixPtr Inverse( MatrixPtr m )
{
	int			i, j, rows = m->rows, cols = m->cols, newRows;
	int			indepEQ = 0;
	double		**a = m->mat;
	double		**u = NULL, **v = NULL, *w = NULL, *vr1 = NULL;
	MatrixPtr	mInv = NULL;
	
	/*	Make sure we have at least as much rows as columns.						*/
	if ( cols > rows ) newRows = cols;
	else newRows = rows;

	u = New2DArray( newRows, cols );
	v = New2DArray( newRows, cols );
	w = New1DArray( newRows );

	for ( i = 0; i < rows; ++i ) {
		for ( j = 0; j < cols; ++j ) {
			u[i][j] = a[i][j];
		}
	}
		
	svdcmp( u, newRows, cols, w, v, NULL );
	
	if ( ( indepEQ = svrank( w, newRows, kTiny ) ) != newRows )	{
		FatalError( "Unexpected singular matrix." );
/*		fputs( "Unexpected singular matrix.", stderr );*/
	}
			
	mInv = NewMatrix( cols, rows, kRowPointers );
	
	svinverse( mInv->mat, cols, u, w, v );

	Free1DArray( w );
	Free2DArray( v );
	Free2DArray( u );
	
	return mInv;
}


