/*
	Matrix.c: Vector/Matrix allocation and deallocation functions.

	RedMech is a tool for automatically reducing CnHm-mechanisms.
	Written by J. Goettgens, P. Terhoeven, R. Korff	
*/

#ifdef applec
#pragma load "RedMech.dump"
#else
#include "RedMechHeaders.h"
#endif


void MatrixPrint( const MatrixPtr m, FILE *fp, char *label )
{
	int		i, j, row = m->rows, col = m->cols;
	double	**a = m->mat;
	
	if ( !fp ) fp = stdout;
	
	if ( gMathematica ) {
		
		fprintf( fp, "m = { " );

		for ( i = 0; i < row; ++i ) {
			for ( j = 0; j < col; ++j ) {
				fprintf( fp, "%s%g", (j==0) ? " { " : ", ", a[i][j] );
			}
			fprintf( fp, "%s", (i == row-1) ? " }\n" : " },\n" );
		}
		fprintf( fp, "}\n" );
	
	} else {
		
		if ( label ) fprintf( fp, "%s:\n", label );
	
		for ( i = 0; i < row; ++i ) {
			for ( j = 0; j < col; ++j ) {
				fprintf( fp, "\t%2g", a[i][j] );
			}
			fprintf( fp, "\n" );
		}
		fprintf( fp, "\n" );

	}
}


static double dotprod( int n, double *x, int incx, double *y, int incy )
{
	int		i;
	double	sum = 0.0;
	
	for ( i = 0; i < n; ++i, x += incx, y += incy ) {
		sum += *x * *y;
	}
	return sum;
}


MatrixPtr Transpose( const MatrixPtr m )
{
	int			i, j, row = m->rows, col = m->cols;
	MatrixPtr	trans = NULL;
	double		**ta = NULL, **ma = NULL;
	
	trans = NewMatrix( col, row, kRowPointers );
	
	ta = trans->mat, ma = m->mat;
	for ( i = 0; i < row; ++i ) {
		for ( j = 0; j < col; ++j ) {
			ta[j][i] = ma[i][j];
		}
	}
	
	return trans;
}


MatrixPtr MatrixMult( const MatrixPtr mat1, const MatrixPtr mat2 )
{
	int			i, j; 
	int			col1 = mat1->cols, col2 = mat2->cols;
	int			row1 = mat1->rows, row2 = mat2->rows;
	MatrixPtr	mul = NULL;
	double		**a1 = NULL, **a2 = NULL, **a = NULL;
	
	if ( col1 != row2 ) 
		FatalError( "Illegal matrix dimensions for multiplication." );
		
	mul = NewMatrix( row1, col2, kRowPointers );
	
	a1 = mat1->mat, a2 = mat2->mat, a = mul->mat;
	for ( i = 0; i < row1; ++i ) {
		for ( j = 0; j < col2; ++j ) {
			a[i][j] = dotprod( col1, a1[i], 1, &a2[0][j], col2 );
		}
	}
	
	return mul;
}


void ClearMatrix( MatrixPtr m )
{
	memset( (void *)m->mat[0], 0, m->phys_rows * m->phys_cols * sizeof(m->mat[0][0]) );
}
