/*
	Reaction.c: Routines that deal with the i/o of the mechanism.

	RedMech is a tool for automatically reducing CnHm-mechanisms.
	Written by J. Goettgens, P. Terhoeven, R. Korff	
*/

#ifdef applec
#pragma load "RedMech.dump"
#else
#include "RedMechHeaders.h"
#endif


ReactionPtr NewReactionArray( int numOfReacs )
{
	ReactionPtr		p = NULL;
	
	p = (ReactionPtr)malloc( numOfReacs * sizeof(Reaction) );
	if ( !p ) FatalError( "NewReactionArray(): Out of memory." );
	
	memset( (char *)p, 0, numOfReacs * sizeof(Reaction) );
	
	return p;
}


void FreeReactionArray( ReactionPtr p )
{
	free( p );
}


ReactionPtr GetReactions( FILE *fp, int numOfReacs, SpeciesListPtr list )
{
	Token			theToken;
	TokenType		lastToken = kInvalidToken;
	int				onRHS = FALSE;
	int				hasThirdBody = FALSE;
	int				notDone = TRUE;
	int				line = 0, lastLine = 0, curReac = 0;
	double			coeff = 0.0, lastNumber = 0.0;
	ReactionPtr		theReaction = NULL, allReactions = NULL;
	char			errorString[256];
	
	memset( (char *)&theToken, 0, sizeof(theToken) );
	allReactions = NewReactionArray( numOfReacs );
	
	do {
		Scanner( &theToken, fp, &line );
		
		switch ( theToken.type ) {
			case kLabel:
				if ( lastToken != kIdentifier || !onRHS ) {
					if ( curReac > 0 ) {
						sprintf( errorString, "unexpected label \"%s\"", theToken.string );
						ParseError( line, errorString );	/* Not first reaction.	*/
					}
				}
				if ( curReac > 0 )
					CheckStoichiometry( &allReactions[curReac-1], list, lastLine );
				lastLine = line;
				theReaction = &allReactions[curReac++];
				LowerString( theToken.string );
				strcpy( theReaction->label, theToken.string );
				onRHS = FALSE;							/* Now on left hand side.	*/
				hasThirdBody = FALSE;
				lastToken = theToken.type;
				break;
			case kPlus:
				if ( lastToken != kIdentifier ) {
					ParseError( line, "unexpected \"+\" sign" );
				}
				lastToken = theToken.type;
				break;
			case kIdentifier:
				if ( lastToken == kIdentifier ) {
					sprintf( errorString, "unexpected identifier \"%s\"", 
						theToken.string );
					ParseError( line, errorString );
				}
				UpperString( theToken.string );
				if ( strcmp( theToken.string, gThirdBody ) == 0 ) {
					if ( onRHS && !hasThirdBody ) {
						ParseError( line, "third body not on left hand side" );
					}
					hasThirdBody = TRUE;
					lastToken = theToken.type;
					break;
				}
				coeff = 1.0;
				if ( lastToken == kNumber ) coeff *= lastNumber;
				if ( !onRHS ) coeff *= -1.0;
				theReaction->stoichCoeff[theReaction->numReactants] = coeff;
				theReaction->speciesIndex[theReaction->numReactants] = 
					Row( theToken.string, list );
				if ( theReaction->speciesIndex[theReaction->numReactants] < 0 )
					ParseError( line, "illegal species found" );
				theReaction->numReactants++;
				lastToken = theToken.type;
				break;
			case kReactionDirection:
				if ( lastToken != kIdentifier || onRHS ) {
					sprintf( errorString, "unexpected \"%s\"", theToken.string );
					ParseError( line, errorString );
				}
				onRHS = TRUE;
				lastToken = kReactionDirection;
				break;
			case kNumber:
				if ( !( lastToken == kReactionDirection || lastToken == kPlus ||
					lastToken == kLabel ) ) {
					sprintf( errorString, "unexpected number \"%g\"", 
						theToken.value );
					ParseError( line, errorString );
				}
				lastNumber = theToken.value;
				lastToken = theToken.type;
				break;
			case kDone:
				notDone = FALSE;		/* don't forget to check last reaction		*/
				CheckStoichiometry( &allReactions[curReac-1], list, lastLine );
				break;
			default:
				sprintf( errorString, "illegal token %d\n", theToken.type );
				ParseError( line, errorString );
				break;
		}
	} while ( notDone );
	
	if ( curReac != numOfReacs ) 
		FatalError( "Inconsistent number of reactions in FillReactions()." );
	
	return allReactions;
}


int Column( char *label, ReactionPtr reacs, int numOfReacs )
{
	int		i;
	
	LowerString( label );
	
	for ( i = 0; i < numOfReacs; ++i ) {
		if ( strcmp( label, reacs[i].label ) == 0 ) return i;
	}
	
	return -1;							/* reaction label not found					*/
}



int AlreadyInList( const ReactionPtr cur, const ReactionPtr list, int n )
{
	int		i, j, k;
	int		this, reactants = cur->numReactants;
	int		inList = 0;
	
	if ( n == 0 ) return inList;
	
	for ( i = 0; i < n; ++i ) {
		if ( reactants != list[i].numReactants ) continue;
		inList = 0;
		for ( j = 0; j < reactants; ++j ) {
			this = cur->speciesIndex[j];
			for ( k = 0; k < reactants; ++k ) {
				if ( this == list[i].speciesIndex[k] ) {
					if ( fabs(cur->stoichCoeff[j] - list[i].stoichCoeff[k]) < kTiny ) {
						++inList;
						break;
					}
				}
			}
		}
		if ( inList == reactants ) return TRUE;
	}
	
	return FALSE;
}


void PrintReactions( ReactionPtr allReacs, SpeciesListPtr list,
	int numOfReacs, FILE *fp, char *label )
{	
	int 	i;
	
	if ( !fp ) fp = stdout;
	
	if ( label ) fprintf( fp, "%s\n", label );

	for ( i = 0; i < numOfReacs; ++i ) {
		/*	Check here for duplicate equations.										*/
		if ( gXtra )
			if ( AlreadyInList( &allReacs[i], allReacs, i ) ) continue;
		PrintOneReaction( &allReacs[i], list, fp );
	}
	fprintf( fp, "\n" );
}


void PrintOneReaction( ReactionPtr reac, SpeciesListPtr list, FILE *fp )
{
	static char			left[256], right[256], temp[256];
	static const char	*dir = " ->";
	int					j, nSpecies;
	int					firstLeft = TRUE, firstRight = TRUE;
	double				coeff;
	
	nSpecies = reac->numReactants;
	if ( nSpecies == 0 ) return;			/* i.e. this reaction was eliminated.	*/

	sprintf( left, "" );
	sprintf( right, "" );
	
	for ( j = 0; j < nSpecies; ++j ) {
#		ifdef applec
		SpinCursor( 1 );
#		endif
		coeff = reac->stoichCoeff[j];
		if ( coeff > kTiny ) {				/* Put positive coefficients on rhs.	*/
			if ( fabs( coeff - 1.0 ) < kTiny ) {
				sprintf( temp, "%s%s", ( firstRight ) ? " " : " + ", 
					SpeciesName( list, reac->speciesIndex[j] ) );
			} else {
				sprintf( temp, "%s%g %s", ( firstRight ) ? " " : " + ", coeff, 
					SpeciesName( list, reac->speciesIndex[j] ) );
			}
			strcat ( right, temp );
			firstRight = FALSE;
		}
		else if ( coeff < -kTiny ) {		/* Put negative coefficients on lhs.	*/
			if ( fabs( coeff + 1.0 ) < kTiny ) {
				sprintf( temp, "%s%s", ( firstLeft ) ? " " : " + ",  
					SpeciesName( list, reac->speciesIndex[j] ) );
			} else {
				sprintf( temp, "%s%g %s", ( firstLeft ) ? " " : " + ", -coeff, 
					SpeciesName( list, reac->speciesIndex[j] ) );
			}
			strcat ( left, temp );
			firstLeft = FALSE;
		}
	}
	if ( !firstLeft && !firstRight ) {
		fprintf( fp, "\t%s:", reac->label );
		fprintf( fp, "%s%s%s\n", left, dir, right );
	}
}

#undef DEBUG


void CheckStoichiometry( ReactionPtr reac, SpeciesListPtr list, int line )
{
	static char		errorString[256];
	int				elemFound = -1, elemCoeff;
	int				i, j, num = reac->numReactants;
	const char		*pos = NULL;
#ifdef mips
	double			elemCount[kElems];
#else
	double			elemCount[kElems] = { 0.0, 0.0, 0.0, 0.0, 0.0 };
#endif
	double			count = 0.0;

#ifdef mips
	for ( i = 0; i < kElems; ++i ) elemCount[i] = 0.0;
#endif

	for ( i = 0; i < num; ++i ) {

		/*	take care of isomeres
		*/
		pos = StripIsomerePrefix( list->species[ reac->speciesIndex[i] ] );

		do {
			elemCoeff = 1;
			if ( isalpha(*pos) ) {
				elemFound = -1;
				for ( j = 0; j < kElems; ++j ) {
					if ( *pos == gTheElems[j] ) {
						count = reac->stoichCoeff[i];
						elemFound = j;
#						ifdef DEBUG
						fprintf( stderr, "# found element: \"%c\"\n", gTheElems[j] );
#						endif
						break;
					}
				}
			}
			if ( elemFound == -1 ) {
				sprintf( errorString, "unrecognized element \"%s\"", 
					list->species[ reac->speciesIndex[i] ] );
				ParseError( line, errorString );
			}
			if ( isdigit( *(pos+1) ) ) {				/* Check for digit.			*/
				elemCoeff = atoi( ++pos );
				while ( isdigit(*pos) ) ++pos;
				--pos;
#				ifdef DEBUG
				fprintf( stderr, "# found element coefficients: \"%d\"\n", elemCoeff );
#				endif
			}
			
			elemCount[elemFound] += count * (double)elemCoeff;
#			ifdef DEBUG
			fprintf( stderr, "# elemCount = %g\n", elemCount[elemFound] );
#			endif
		} while ( *(++pos) );
	}

	for ( i = 0; i < kElems; ++i ) {
		if ( fabs( elemCount[i] ) > kTiny ) {
			fprintf( stderr, "#\t" );
			PrintOneReaction( reac, list, stderr );
			sprintf( errorString, "wrong stoichiometry for element \"%c\" detected", 
				gTheElems[i] );
			ParseError( line, errorString );
		}
	}
}


