/*
	RedMech.c: Main program of RedMech and functions related to the 
			  reduction process.
	
	RedMech is a tool for automatically reducing CnHm-mechanisms.
	Written by J. Goettgens, P. Terhoeven, R. Korff	
	
	History:	11/10/94	Bug fix for a single global reaction
							(thanks to Christos Frouzakis)
*/

#ifdef applec
#pragma load "RedMech.dump"
#else
#include "RedMechHeaders.h"
#endif

const double	kTiny = 1.0E-9;
const int		kLineLength = 60;
char			*gReactionFile = NULL;
char			*gReductionFile = NULL;
char			*gFileName = NULL;
char			*gBuffer = NULL;
char			*gGlobalReactionFile = NULL;
int				gProgress = FALSE;
int				gXtra = TRUE;
int				gGlobal = TRUE;
int				gMathematica = FALSE;
int				gAuto = FALSE;


ReactionPtr AutoGlobalReactions( int numGlobal, int numSpecies,
		ReactionPtr reduced, int numOfReduced );


int main( int argc, char *argv[] )
{
	SpeciesListPtr	speciesList = NULL;
	ReactionPtr		reacs = NULL, reduced = NULL;
	ReactionPtr		globalReacs = NULL;
	RedRecPtr		redList = NULL;
	ConsEqPtr		cons = NULL;
	FILE			*fp = NULL;
	int				numOfReacs = 0, numOfReduced = 0, numOfReducs = 0;
	int				elems = 0, numGlobal = 0, rank = 0;
	MatrixPtr		m = NULL;

	HandleCommandLine( argc, argv );

	/*	Allocate a global buffer for output.
	*/
	if ( !( gBuffer = (char *)malloc( kBufSize * sizeof(*gBuffer) ) ) )
		FatalError( "memory exhausted." );
	memset( gBuffer, 0, kBufSize * sizeof(*gBuffer) );

	if ( !(fp = fopen( gReactionFile, "r" )) )
		FatalError( "Couldn't open reaction file." );

	/*	Collect all species.													*/
	speciesList = AssembleSpeciesList( fp, &numOfReacs );
	elems = CountElements( speciesList );
	if ( gProgress ) {
		PrintSpeciesList( speciesList, numOfReacs, stdout, NULL );
		fprintf( stdout, "%d different elements found.\n\n", elems );
		fflush( stdout );
	}
	rewind( fp );
	
	/*	Read in the mechanism.													*/
	reacs = GetReactions( fp, numOfReacs, speciesList );
	fclose( fp );								/*	We no longer need this guy.	*/
	PrintReactions( reacs, speciesList, numOfReacs, stdout, "Original mechanism:" );
	
	/*	Assemble the matrix of stoichiometric coefficients.						*/
	m = AssembleMatrix( numOfReacs, speciesList, reacs );
	rank = IndependentEQ( m );
	if ( gProgress > 1 ) {
		MatrixPrint( m, stdout, "Stoichiometric coefficient matrix (original mechanism)" );
		fprintf( stdout, "Rank is %d.\n\n", rank );
		fflush( stdout );
	}
	
	/*	Read the steady state assumptions.										*/
	if ( !( fp = fopen( gReductionFile, "r" ) ) )
		FatalError( "Couldn't open reduction file\n" );
	gFileName = gReductionFile;			/* Set up file name for error messages	*/
	redList = GetReductions( fp, numOfReacs, speciesList, reacs, m, &numOfReducs );
	// fclose( fp );								/*	No longer needed.			*/
	if ( gProgress ) PrintReductions( redList, stdout, NULL );
	
	numOfReduced = numOfReacs - numOfReducs;
	/*numGlobal = speciesList->num - numOfReducs - elems; wrong!!! (jg) */
	numGlobal = rank - numOfReducs;
	fprintf( stdout, "There are %d linearly independent global reactions.\n", 
			 numGlobal );
	if ( speciesList->num - numOfReducs - elems > numGlobal ) {
		fprintf( stdout, "The maximum possible number is %d.\n", 
			speciesList->num - numOfReducs - elems );
	}
	putc('\n', stdout ); fflush( stdout );

	cons = ConsEqFromReductions( redList, speciesList, m );
	PrintConsEq( cons, speciesList->num, numOfReacs, reacs, stdout, "original mechanism" );
	
	/*	Here's the beef.  Reduce the mechanism.									*/
	Reduce( m, redList );

	if ( gProgress > 1 ) 
		MatrixPrint( m, stdout, "Stoichiometric coefficient matrix (reduced mechanism):" );
		fflush( stdout );
	
	/*	List all non steady-state reactions.									*/
	reduced = CollectReducedReactions( cons, speciesList, numOfReacs, reacs, 
									   redList, &numOfReduced );

	/*	Print species balance equations for all non steady-state species.		*/
	PrintConsEq( cons, speciesList->num, numOfReacs, reacs, stdout, "reduced mechanism" );
	fflush( stdout );
	PrintReactions( reduced, speciesList, numOfReduced, stdout, "Reduced mechanism:" );
	fflush( stdout );
	
	if ( gGlobal ) {
		int			numUserGlobals = numGlobal;
		fflush( stdout );
		/*	Let the (expert) user choose her favorite global reactions.			*/
		if ( gGlobalReactionFile ) {
			SpeciesListPtr	globalSpeciesList = NULL;
			
			gFileName = gGlobalReactionFile;
			if ( !(fp = fopen( gGlobalReactionFile, "r" )) )
				FatalError( "Couldn't open reaction file." );
			globalSpeciesList = AssembleSpeciesList( fp, &numUserGlobals );
			rewind( fp );
			globalReacs = GetReactions( fp, numUserGlobals, speciesList );
			fclose( fp );
			PrintReactions( globalReacs, speciesList, numUserGlobals, stdout, "Global reactions read from user file:" );
			FreeSpeciesList( globalSpeciesList );
		} else {
			if ( gAuto ) {
				globalReacs = AutoGlobalReactions( numGlobal, speciesList->num, 
					reduced, numOfReduced );
			} else {
				globalReacs = GlobalReactions( numGlobal, speciesList->num, 
					reduced, numOfReduced );
			}
			PrintReactions( globalReacs, speciesList, numGlobal, stdout, "Global reactions:" );
		}
		fflush( stdout );
		ComputeGlobalRates( globalReacs, numUserGlobals, reduced, numOfReduced, 
			speciesList, stdout );
		fflush( stdout );
		FreeReactionArray( globalReacs );
	}

	/*	Clean up.																*/
	DisposeMatrix( m );
	FreeSpeciesList( speciesList );
	FreeReactionArray( reacs );
	FreeReactionArray( reduced );
	FreeReductions( redList );
	free( gBuffer );
	
	fflush( stdout );
	fputs( "\n# done.\n", stderr );
	return 0;									/* ... to make the shell happy	*/
}


void Reduce( MatrixPtr m, RedRecPtr redList )
{
	int			rows = m->rows, cols = m->cols;
	int			rowR, colR, row, col;
	double		factor;
	double		**a = m->mat;
	RedRecPtr	rr = redList;
	
	/*	Eliminate the rates of the reactions that were specified together 
		with the steady state assumptions.										*/
	while ( rr ) {	
		rowR = rr->row;
		colR = rr->column;
		/*	Check for spurious zero pivots (jg, 30.7.92).						*/
		if ( fabs(a[rowR][colR]) < kTiny )
			FatalError( "Zero pivot element encountered" );
		for ( row = 0; row < rows; ++row ) {
			if ( ( row == rowR ) || ( fabs( a[row][colR] ) < kTiny ) ) continue;
			factor = -a[row][colR] / a[rowR][colR];
			for ( col = 0; col < cols; ++col)
				a[row][col] += factor * a[rowR][col];
		}
		rr = rr->next;			/*	Get next reduction record from linked list.	*/
	}
}



MatrixPtr AssembleMatrix( int numOfReacs, SpeciesListPtr sl, ReactionPtr reacs )
{
	int			i, j, k;
	MatrixPtr	m = NULL;
	double		**a = NULL;

	m = NewMatrix( sl->num, numOfReacs, kRowPointers );
	a = m->mat;
	
	for ( j = 0; j < numOfReacs; ++j ) {
		for( i = 0; i < reacs[j].numReactants; ++i ) {
			k = reacs[j].speciesIndex[i];
			a[k][j] += reacs[j].stoichCoeff[i];
		}
	}

	return m;
}


ReactionPtr CollectReducedReactions( ConsEqPtr cons, SpeciesListPtr list, 
		int numOfReacs, ReactionPtr reacs, RedRecPtr redList, int *numOfReduced )
{
	int			i, j, num, numOfSpecies = list->num;
	ReactionPtr	reduced = NULL, tempRed = NULL;
	RedRecPtr	first = redList;
	Reaction	theReaction;
	int			isReduced;
	int			globalCount;
			
	tempRed = NewReactionArray( *numOfReduced );
	
	/*	Loop over all reactions.													*/
	for ( j = 0, globalCount = 0; j < numOfReacs; ++j ) {
		isReduced = FALSE;
		first = redList;
		while ( first ) {		/* Find out if this reaction has been eliminated.	*/
			if ( first->column == j ) {
				isReduced = TRUE; 
				break;
			}
			first = first->next;
		}
		if ( isReduced ) continue;						/* Skip to next reaction.	*/

		memset( (char *)&theReaction, 0, sizeof(theReaction) );
		sprintf( theReaction.label, "%s", reacs[j].label );

		/*	Loop over all species.													*/
		for ( i = 0; i < numOfSpecies; ++i ) {
			/*	Skip steady state species.											*/
			if ( cons[i].isSteadyState || fabs( cons[i].row[j] ) < kTiny ) continue;
			num = theReaction.numReactants;
			theReaction.speciesIndex[num] = i;
			theReaction.stoichCoeff[num] = cons[i].row[j];
			++theReaction.numReactants;
		}
		if ( theReaction.numReactants == 0 ) continue;	/* No species in reaction.	*/
		
		if ( globalCount == *numOfReduced )
			FatalError( "Too many reduced reactions." );

		tempRed[globalCount] = theReaction;
		++globalCount;
	}
	
	if ( globalCount < *numOfReduced ) {
		*numOfReduced = globalCount;
		reduced = NewReactionArray( *numOfReduced );
		memcpy( (char *)reduced, (char *)tempRed, *numOfReduced * sizeof(Reaction) );
		FreeReactionArray( tempRed );
	} else {
		reduced = tempRed;
	}
	
	return reduced;
}


ReactionPtr GlobalReactions( int numGlobal, int numSpecies,
		ReactionPtr reduced, int numOfReduced )
{
	int				i, j, k, n, gc;
	int				found, duplicLabel, globCount, done = FALSE;
	int				bufferSize = 256;
	char			*buffer = NULL;
	char			*bp = NULL, *label = NULL;
	ReactionPtr		global = NULL;
	MatrixPtr		m = NULL;
	
	global = NewReactionArray( numGlobal );
	m = NewMatrix( numGlobal, numSpecies, kRowPointers );
	buffer = (char *)malloc( bufferSize * sizeof(*buffer) );
	if ( !buffer ) FatalError( "Storage exhausted" );
	*buffer = '\0';
	
	fflush( stdout );
	
	/*  Prompt the user to choose global reactions.                                 */
	while ( !done ) {
	
		fprintf( stderr, "Choose %d global reactions:\n", numGlobal );
		
		if ( fgets( buffer, bufferSize-1, stdin ) == NULL ) 
			FatalError( "Couldn't read input line (GlobalReactions)" );
		
		if ( *buffer && buffer[strlen(buffer)-1] != '\n' ) {
			fputs( "# input buffer too short, please reenter global reactions...\n", stderr );
			free( buffer );
			fflush(stdin);
			bufferSize *= 2;
			buffer = (char *)malloc( bufferSize * sizeof(*buffer) );
			if ( !buffer ) FatalError( "Storage exhausted" );
			*buffer = '\0';
			continue;
		}
		bp = buffer;
		globCount = 0;
		ClearMatrix( m );
		
		while ( isspace( *bp ) ) ++bp;				/*	Go to first reaction.		*/
		for ( gc = 0; gc < numGlobal; ++gc ) {

			if ( !bp ) break;						/*	End of string encountered.	*/
			if ( (sscanf( bp, "%s", global[gc].label )) != 1 ) {
				fprintf( stderr, "# not enough reactions were specified.\n" );
				break;
			}
			++globCount;
			
			found = FALSE; duplicLabel = FALSE;
			for ( j = 0; j < numOfReduced; ++j ) {
				label = reduced[j].label;
				if ( strcmp( global[gc].label, label ) == 0 ) {
					for ( k = 0; k < gc; ++k ) {
						if ( strcmp( global[k].label, label ) == 0 ) {
							fprintf( stderr, "# duplicate label \"%s\" found.\n", label );
							duplicLabel = TRUE;
							break;
						}
					}
					if ( duplicLabel ) break;
					global[gc] = reduced[j];
					found = TRUE;
					break;
				}
			}
			if ( duplicLabel ) {
				break;
			}
			if ( !found ) {
				fprintf( stderr, "# Reaction \"%s\" not found.\n", global[gc].label );
				break;
			}
						
			/*	Check for linear independence.										*/
			if ( gc > 0 ) {
				for ( i = 0; i < globCount; ++i ) {
					n = global[i].numReactants;
					for ( j = 0; j < n; ++j ) {
						k = global[i].speciesIndex[j];
						m->mat[i][k] = global[i].stoichCoeff[j];
					}
				}
				if ( IndependentEQ( m ) != globCount ) {
					fprintf( stderr, "# Resulting equations are linearly dependent.\n" );
					fprintf( stderr, "# Offending label is \"%s\".\n", label );
					break;
				}
			}
			done = (globCount == numGlobal);		/* fixed 11/10/94				*/
			
			while ( isalnum( *bp ) ) ++bp;			/*	Advance to next reaction	*/
			while ( isspace( *bp ) ) ++bp;
			if ( done && *bp )  {
				fputs( "# too many reactions, ignoring \"", stderr );
				while ( *bp && *bp != '\n' ) fputc( *bp++, stderr );
				fputs( "\".\n", stderr );
			}
			
		}	/*	for ( gc = 0; gc < numGlobal; ++gc ) ...							*/

	}
	
	for ( i = 0; i < numGlobal; ++i ) {
		RomanNumeral( i+1, global[i].label, TRUE );
	}
	
	DisposeMatrix( m );
	
	return global;
}

ReactionPtr AutoGlobalReactions( int numGlobal, int numSpecies,
		ReactionPtr reduced, int numOfReduced )
{
	int				i, j, k, n, indep;
	int				globCount = 0;
	int				curReac = 0;		/*	current reaction	*/
	char			*label = NULL;
	ReactionPtr		global = NULL;
	MatrixPtr		m = NULL;
	
	globCount = 0;
	
	fprintf( stderr, "Choosing %d global reactions automatically (this may take some time)...\n", numGlobal );
	fprintf( stderr, "global reactions: \"%s", reduced[0].label );
	
	global = NewReactionArray( numGlobal );
	global[0] = reduced[0];
	RomanNumeral( 0+1, global[0].label, TRUE );
	
	if ( numGlobal > 1 ) {
		globCount = 1; curReac = 1;
		m = NewMatrix( numGlobal, numSpecies, kRowPointers );
		for ( curReac = 1; curReac < numOfReduced; ++curReac ) {
		
			global[globCount] = reduced[curReac];
			RomanNumeral( globCount+1, global[globCount].label, TRUE );
			++globCount;
			
			/*	Check for linear independence.										*/
			ClearMatrix( m );
			for ( i = 0; i < globCount; ++i ) {
				n = global[i].numReactants;
				for ( j = 0; j < n; ++j ) {
					k = global[i].speciesIndex[j];
					m->mat[i][k] = global[i].stoichCoeff[j];
				}
			}
			if ( ( indep = IndependentEQ( m ) ) == globCount ) {
				/*	we found a global reaction	*/
				fprintf( stderr, " %s", reduced[curReac].label ); fflush( stderr );
			} else {
				--globCount;		/*	reset the global reaction counter	*/
			}
			if ( indep == numGlobal ) {
				fputs( "\"\n\n", stderr );
				break;
			} 	
		}	/*	for ( curReac = 1; curReac < numReduced; ++curReac ) ...							*/
		DisposeMatrix( m );
	}
	
	return global;
}


void ComputeGlobalRates( ReactionPtr global, int numGlobal, ReactionPtr reduced, 
		int numOfReduced, SpeciesListPtr specList, FILE *fp )
{
	int				i, j, k, n, numSpecs = specList->num;
	int				specCount = 0, redCount = 0;
	int				*list = NULL;
	MatrixPtr		bm = NULL, x = NULL, rhs = NULL, tm = NULL, mul = NULL;
	MatrixPtr		tmx = NULL, muli = NULL, m = NULL;
	char			roman[16];
	char			*bptr = gBuffer;
	
	if ( fp == NULL ) fp = stdout;
	
	/*	Create a list to map the indices of the remaining species onto a 
		series to index a matrix.													*/
	list = (int *)malloc( numSpecs * sizeof( int) );
	if ( !list ) FatalError( "Out of juice in ComputeGlobalRates()." );
	memset( (char *)list, 0, numSpecs * sizeof(int) );
	
	for ( i = 0; i < numOfReduced; ++i ) {
		n = reduced[i].numReactants;
		for ( j = 0; j < n; ++j ) {
			k = reduced[i].speciesIndex[j];
			++list[k];
		}
	}
	
	for ( i = 0; i < numSpecs; ++i ) {
		if ( list[i] ) list[i] = specCount++;
		else list[i] = -1;
/*		fprintf( stderr, "\"%s\": %d -> %d\n", specList->species[i], i, list[i] );*/
	}
	
	/*	Store the stoichiometric coefficients of the global reaction's species
		in the columns of the matrix bm.											*/
	bm = NewMatrix( specCount, numGlobal, kRowPointers );
	for ( i = 0; i < numGlobal; ++i ) {
		n = global[i].numReactants;
		for ( j = 0; j < n; ++j ) {
			k = global[i].speciesIndex[j];
			if ( list[k] < 0 ) {
				fprintf( stderr, "### FATAL ERROR: Species \"%s\" is steady state\n", specList->species[k], 
					k, list[k] );
				exit( 2 );
			}
			bm->mat[ list[k] ][i] += global[i].stoichCoeff[j];
		}
	}
		
	/*	Interpret the rows of matrix bm as species balance equations.				*/
	fputs( "Species balance equations:\n", fp );
	for ( i = 0; i < bm->rows; ++i ) {
		int		theSpec = -1;
		int		first;
		
		for ( j = 0; j < numSpecs; ++j ) {
			if ( i == list[j] ) {
				theSpec = j;			/* Get the species index of the matrix row.	*/
				break;
			}
		}
		if ( theSpec == -1 ) FatalError( "Species not found.\n");
		
		/*bptr +=*/ sprintf( bptr, "\tL([%s]) = ", SpeciesName( specList, theSpec ) );
		bptr += strlen( bptr );
		first = TRUE;
		for ( j = 0; j < bm->cols; ++j ) {
			double	number;
			
			RomanNumeral( j+1, roman, TRUE );
			
			if ( fabs( number = bm->mat[i][j] ) > kTiny ) {
				if ( bptr - gBuffer > kLineLength ) {
					fprintf( fp, "%s\n\t\t", bptr = gBuffer );
				}
				if ( number < 0.0 ) {
					if ( fabs( number + 1.0 ) < kTiny ) {
						/*bptr +=*/ sprintf( bptr, "- w%s ", roman );
						bptr += strlen(bptr);
					} else {
						/*bptr +=*/ sprintf( bptr, "- %g w%s ", -number, roman );
						bptr += strlen(bptr);
					}
					first = FALSE;
				}
				else {
					if ( fabs( number - 1.0 ) < kTiny ) {
						/*bptr +=*/ sprintf( bptr, "%sw%s ", (first) ? "" : "+ ", roman );
						bptr += strlen(bptr);
					} else {
						/*bptr +=*/ sprintf( bptr, "%s%g w%s ", (first) ? "" : "+ ", number, roman );
						bptr += strlen(bptr);
					}
					first = FALSE;
				}
			}
		}
		fprintf( fp, "%s\n", bptr = gBuffer );
	}
	fputs( "\n", fp );

/*	MatrixPrint( bm, stdout, "global matrix" );*/

	x = NewMatrix( specCount, numOfReduced, kRowPointers );
	for ( i = 0; i < numOfReduced; ++i ) {
		n = reduced[i].numReactants;
		for ( j = 0; j < n; ++j ) {
			k = reduced[i].speciesIndex[j];
			if ( list[k] < 0 ) {
				fprintf( stderr, "### FATAL ERROR: Species \"%s\" is steady state\n", specList->species[k], 
					k, list[k] );
				exit( 2 );
			}
			x->mat[ list[k] ][i] = reduced[i].stoichCoeff[j];
		}
	}
/*	MatrixPrint( x, stdout, "rhs matrix" );*/
	
	/*	One of these days, we will clean up this mess...						*/
	
	tm = Transpose( bm );
/*	MatrixPrint( tm, stdout, "transposed global matrix" );*/

	tmx = MatrixMult( tm, x );
/*	MatrixPrint( tmx, stdout, "transposed . rhs" );*/
	
	mul = MatrixMult( tm, bm );
/*	MatrixPrint( mul, stdout, "transposed . global" );*/
	
	muli = Inverse( mul );
/*	MatrixPrint( muli, stdout, "inverse" );*/
		
	m = MatrixMult( muli, mul );
/*	MatrixPrint( m, stdout, "inverse . matrix" );*/
	
	m = MatrixMult( muli, tmx );
/*	MatrixPrint( m, stdout, "final" );*/
	
	/*	the rows of m are the rate coefficients for the computation of the global
		rates
	*/
	fputs( "Global rates:\n", fp );
	for ( i = 0; i < m->rows; ++i ) {
		int 	first;
		RomanNumeral( i+1, roman, TRUE );
		/*bptr +=*/ sprintf( bptr, "\tw%s = ", roman );
		bptr += strlen(bptr);
		first = TRUE;
		for ( j = 0; j < m->cols; ++j ) {
			double	number;
			
			if ( fabs( number = m->mat[i][j] ) > kTiny ) {
				if ( bptr - gBuffer > kLineLength ) {
					fprintf( fp, "%s\n\t\t", bptr = gBuffer );
				}
				
				if ( number < 0.0 ) {
					if ( fabs( number + 1.0 ) < kTiny ) {
						/*bptr +=*/ sprintf( bptr, "- w%s ", reduced[j].label );
						bptr += strlen(bptr);
					} else {
						/*bptr +=*/ sprintf( bptr, "- %g w%s ", -number, reduced[j].label );
						bptr += strlen(bptr);
					}
					first = FALSE;
				}
				else {
					if ( fabs( number - 1.0 ) < kTiny ) {
						/*bptr +=*/ sprintf( bptr, "%sw%s ", (first) ? "": "+ ", reduced[j].label );
						bptr += strlen(bptr);
					} else {
						/*bptr +=*/ sprintf( bptr, "%s%g w%s ", (first) ? "" : "+ ", number, 
							reduced[j].label );
						bptr += strlen(bptr);
					}
					first = FALSE;
				}
			}
		}
		fprintf( fp, "%s\n", bptr = gBuffer );
	}
	fprintf( fp, "\n" );

	DisposeMatrix( tmx );					/*	Don't be a memory hog.				*/
	DisposeMatrix( tm );
	DisposeMatrix( m );
	DisposeMatrix( mul );
	DisposeMatrix( muli );
	DisposeMatrix( x );
	DisposeMatrix( bm );
	free( list );
}
