/*
	RedMech.h: Header file for the RedMech tool.
	
	RedMech is a tool for automatically reducing CnHm-mechanisms.
	Written by J. Goettgens, P. Terhoeven, R. Korff
	
	Latest Revision: 04/24/92
*/

#ifndef __RedMech__
#define __RedMech__

#include <stdio.h>

#ifdef applec
#pragma once
#endif

#ifndef TRUE
#define TRUE	1
#define FALSE	0
#endif

/*	kMaxLen  is used for the length of various strings.							*/
#define kMaxLen	40

#include "Arrays.h"


/* 
	Scanner stuff
*/

typedef enum {
	kInvalidToken = -1,
	kMinus		  = '-',
	kPlus		  = '+',
	kLabel		  = 256,
	kIdentifier,
	kReactionDirection,
	kNumber,
	kDone
} TokenType;

typedef struct Token {
	TokenType	type;
	char		string[kMaxLen];
	double		value;
} Token, *TokenPtr;


int Scanner( TokenPtr tok, FILE *fp, int *inLine );


/*
	SpeciesList stuff
*/

/*	kMaxSpecies  specifies the maximum number of species we can handle.			*/


#define kMaxSpecies	150
#define kElems		5					/* chemical elements (incl. third body)	*/

typedef struct SpeciesList {
	int		num;						/* number of species					*/
	char	*species[kMaxSpecies];		/* species names						*/
	char	*names;						/* memory pool for names				*/
	char	*cur;						/* current free position				*/
} SpeciesList, *SpeciesListPtr;

SpeciesListPtr AssembleSpeciesList( FILE *fp, int *numOfReacs );
void FreeSpeciesList( SpeciesListPtr p );
int Row( char *speciesName, SpeciesListPtr list );
char *SpeciesName( SpeciesListPtr list, int index );
void PrintSpeciesList( SpeciesListPtr list, int numOfReacs, 
						FILE *fp, char *label );
int CountElements( SpeciesListPtr list );


/*
	Reaction stuff
*/

/*	Maximum number of reactants we can handle.									*/
#define kMaxReactants	20

typedef struct Reaction {
	char	label[kMaxLen];
	int		numReactants;
	int		speciesIndex[kMaxReactants];
	double	stoichCoeff[kMaxReactants];
} Reaction, *ReactionPtr;

ReactionPtr NewReactionArray( int numOfReacs );
void FreeReactionArray( ReactionPtr p );
ReactionPtr GetReactions( FILE *fp, int numOfReacs, SpeciesListPtr list );
int Column( char *label, ReactionPtr reacs, int numOfReacs );
void PrintReactions( ReactionPtr allReacs, SpeciesListPtr list,
					int numOfReacs, FILE *fp, char *label );
void PrintOneReaction( ReactionPtr reac, SpeciesListPtr list, FILE *fp );
void CheckStoichiometry( ReactionPtr reac, SpeciesListPtr list, int line );


/*
	Matrix stuff
*/
void MatrixPrint( const MatrixPtr m, FILE *fp, char *label );
MatrixPtr Transpose( const MatrixPtr m );
MatrixPtr MatrixMult( const MatrixPtr mat1, const MatrixPtr mat2 );
void ClearMatrix( MatrixPtr m );

/*
	Reduction stuff
*/

/*	Size of text buffer in GetReductions().										*/
#define kBufSize	255

typedef struct ReductionRecord {
	char			species[kMaxLen];
	int				row;
	char			label[kMaxLen];
	int				column;
	struct ReductionRecord *next;
} ReductionRecord, *RedRecPtr;

typedef struct ConsEq {
	char	*speciesName;				/* name of species						*/
	int		isSteadyState;				/* species is steady state				*/
	double	*row;						/* row of coefficient matrix			*/
} ConsEq, *ConsEqPtr;

RedRecPtr NewReduction( void );
RedRecPtr GetReductions( FILE *fp, int numOfReacs, SpeciesListPtr sl,
						ReactionPtr reacs, MatrixPtr m, int *numOfReducs);
void FreeReductions( RedRecPtr redList );
void PrintReductions( RedRecPtr redList, FILE *fp, char *label );
ConsEqPtr NewConsEqArray( int n );
ConsEqPtr ConsEqFromReductions( RedRecPtr redList, SpeciesListPtr list, 
								MatrixPtr m );
void FreeConsEq( ConsEqPtr cons );
void PrintConsEq( ConsEqPtr cons, int numOfConsEq, int numOfReacs, 
				   ReactionPtr reacs, FILE *fp, char *label );

/*
	IndependentEQ stuff
*/

int IndependentEQ( MatrixPtr m );
MatrixPtr Inverse( MatrixPtr m );

/*
	RedMech stuff
*/

void Reduce( MatrixPtr m, RedRecPtr redList );
MatrixPtr AssembleMatrix( int numOfReacs, SpeciesListPtr sl, ReactionPtr reacs );
ReactionPtr CollectReducedReactions( ConsEqPtr cons, SpeciesListPtr list, 
		int numOfReacs, ReactionPtr reacs, RedRecPtr redList, int *numOfGlobal );
ReactionPtr GlobalReactions( int numGlobal, int numSpecies, ReactionPtr reduced, 
		int numReduced );
void ComputeGlobalRates( ReactionPtr global, int numGlobal, ReactionPtr reduced, 
		int numOfReduced, SpeciesListPtr specList, FILE *fp );


/*	Utility functions															*/
void FatalError( const char *errorString );
void HandleCommandLine( int argc, char *argv[] );
void UpperString( char *s );
void LowerString( char *s );
void ParseError( int line, char *errorString );
void RomanNumeral( int number, char *roman, int caps );
int EmptyLine( const char *buffer );
const char *StripIsomerePrefix( const char *s );
void DoHelp( void );


/*	GLOBALS																		*/
extern char			*gFileName;			/* file name for error messages			*/
extern char 		*gReductionFile;
extern char			*gReactionFile;
extern const char	*gThirdBody;		/* name of third body species			*/
extern const double kTiny;				/* a small number						*/
extern int			gProgress;			/* signals whether to print diagnostics	*/
extern const char	gTheElems[kElems];	/* element names						*/
extern int			gXtra;				/* Experimental stuff...				*/
extern int			gGlobal;			/* handle global reactions				*/
extern char			*gBuffer;			/* line buffer for output				*/
extern const int	kLineLength;
extern int			gMathematica;		/* print matrices in Mathematica form	*/
extern int			gAuto;				/* collect globals automatically	*/
extern char			*gGlobalReactionFile;
#endif	/* __RedMech__ */
