/*
	RedMechUtil.c: General utility functions.

	RedMech is a tool for automatically reducing CnHm-mechanisms.
	Written by J. Goettgens, P. Terhoeven, R. Korff	
*/

#ifdef applec
#pragma load "RedMech.dump"

#include "Resources.h"

#else
#include "RedMechHeaders.h"
#endif

static void ErrMsg( const char *, const char *, int );
static int getopt( int, char *[], const char * );


static int		opterr = 1;
static int		optind = 1;
static int		optopt;
static char	*optarg;
static char 	*gProgName = NULL;


#define ERR(s,c) if(opterr) ErrMsg(argv[0],s,c)

static void ErrMsg( const char *prog, const char *s, int c )
{
	fprintf( stderr, "### %s: %s -- '%c'\n", prog, s, c );
}


static int getopt(int argc, char *argv[], const char *opts)
{
	static int sp = 1;
	register int c;
	register char *cp;

	if ( sp == 1 )
		if ( optind >= argc || argv[optind][0] != '-' || argv[optind][1] == '\0' )
			return(EOF);
		else if ( strcmp(argv[optind], "--") == 0 ) {
			optind++;
			return(EOF);
		}
	optopt = c = argv[optind][sp];
	if ( c == ':' || (cp = strchr(opts, c)) == NULL ) {
		ERR("illegal option", c);
		if ( argv[optind][++sp] == '\0' ) {
			optind++;
			sp = 1;
		}
		return('?');
	}
	if ( *++cp == ':' ) {
		if ( argv[optind][sp + 1] != '\0' )
			optarg = &argv[optind++][sp + 1];
		else if ( ++optind >= argc ) {
			ERR("option requires an argument", c);
			sp = 1;
			return('?');
		} else
			optarg = argv[optind++];
		sp = 1;
	} else {
		if ( argv[optind][++sp] == '\0' ) {
			sp = 1;
			++optind;
		}
		optarg = NULL;
	}
	return(c);
}



void FatalError( const char *errorString )
{
	fprintf( stderr, "###  %s\n", errorString );
	exit( 2 );
}

static void Usage( void )
{
	fprintf( stderr, "# Usage: %s -h      #  for real help\n", gProgName );
	fprintf( stderr, "#    or\n" );
	fprintf( stderr,
"#        %s -m <mechanism> -s <steady states> [-p] [-g] [-a] [-G <global reactions>]\n#\n",
		gProgName );
}


void HandleCommandLine( int argc, char *argv[] )
{
	int		gotReaction = FALSE, gotReduction = FALSE;
	int		c;
	
	/*	Make the program name visible to Usage().											*/
	gProgName = argv[0];

#ifdef qDetailledInfo
	fprintf( stdout,
		"#\n# %s v.1.3 - Another essential utility from the Reducer's Toolkit.\n",
		argv[0] );
	fputs( "# Written by J. Goettgens, P. Terhoeven, R. Korff\n#\n", stdout );
	fflush( stdout );
#else
	fprintf( stderr, "\n# %s v.1.3\n\n", argv[0] );
#endif
	
	while ( (c = getopt(argc, argv, "hpxgzaG:m:s:") ) != EOF) {
		switch( c ) {
			case 'm':
				gReactionFile = optarg;	/* File containing reactions.					*/
				gotReaction = TRUE;
				break;
			case 's':
				gReductionFile = optarg;	/* File containing steady state assumptions.	*/
				gotReduction = TRUE;
				break;
			case 'p':
				gProgress = 3;				/* Additional progress info stuff.				*/
				break;
			case 'h':
				DoHelp();					/* Show help									*/
				exit( 0 );
				break;
			case 'x':
				gXtra = !gXtra;
				break;
			case 'g':
				gGlobal = !gGlobal;
				break;
			case 'z':
				gMathematica = !gMathematica;
				break;
			case 'a':
				gAuto = !gAuto;
				break;
			case 'G':
				gGlobalReactionFile = optarg;
				break;
			case '?':
				Usage();
				exit( 1 );
		}
	}
	if ( !gotReaction ) Usage(), FatalError( "No mechanism file specified." );
	if ( !gotReduction ) Usage(), FatalError( "No steady state assumptions file specified." );
	
	if ( gProgress )
		fprintf( stderr, "# Mechanism: \"%s\"\n# Steady-state assumptions: \"%s\"\n",
				gReactionFile, gReductionFile );
	gFileName = gReactionFile;				/* File name used by Scanner().					*/
}


void UpperString( char *s )
{
	while ( *s ) {
		*s = toupper( *s );
		++s;
	}
}


void LowerString( char *s )
{
	while ( *s ) {
		*s = tolower( *s );
		++s;
	}
}


void ParseError( int line, char *errorString )
{
	fprintf( stderr, "### Parse Error\n\tFile \"%s\"; Line %d # %s\n", 
		gFileName, line, errorString );
	exit( 2 );
}


/*
	Writes number as roman numeral into the string roman, which must be
	large enough to hold the resulting string. Legal input values range
	from 1 to 3999. If numbers fall out of the range, the empty string ""
	is returned.
*/
void RomanNumeral( int number, char *roman, int caps )
{
	int			i, j, len, digit, order = 0;
	char		*ptr = roman;
	char		string[5];
	static const char	letter[4][2] = { {'I', 'V'}, {'X', 'L'}, {'C', 'D'}, {'M', '0'} };

	if ( number < 1 || number > 3999 ) {
		*ptr = '\0';
		return;
	}
	
	sprintf( string, "%d", number );
	len = strlen( string );

	order = len-1;

	for ( i = 0; i < len; ++i ) {						/* Start with the first digit.		*/
		digit = string[i];

		switch ( digit ) {
			case '1':
			case '2':
			case '3':
				for ( j = 0; j < digit - '0'; ++j ) {
					*ptr++ = letter[order][0];
				}
				break;
			case '4':
				*ptr++ = letter[order][0];
				*ptr++ = letter[order][1];
				break;
			case '5':
				*ptr++ = letter[order][1];
				break;
			case '6':
			case '7':
			case '8':
				*ptr++ = letter[order][1];
				for ( j = 0; j < digit - '5'; ++j ) {
					*ptr++ = letter[order][0];
				}
				break;
			case '9':
				*ptr++ = letter[order][0];
				*ptr++ = letter[order+1][0];
				break;
			case '0':
				break;
		}
		--order;
	}
	*ptr = '\0';										/* Terminate string.				*/

	if ( !caps ) LowerString( roman );
}


int EmptyLine( const char *buffer )
{
	do {
		switch( *buffer ) {
			case ' ':
			case '\t':
				/* skip white space		*/
				break;
			case '\n':
			case '\0':
				/* end of line			*/
				return TRUE;
			case '#':
				/* comment line			*/
				return TRUE;
			default:
				/* line contains input	*/
				return FALSE;
		}
	} while ( *buffer++ );
}


const char *StripIsomerePrefix( const char *s )
{
	const char	*p = s + strlen(s);
	
	while ( p >= s ) {
		if ( *p == '-' ) return p+1;	/* return first char after the '-'			*/
		--p;
	}

	return s;							/* no isomere, return start of identifier	*/
}


#ifdef applec

void DoHelp( void )
{
	const int		kHelpText = 400;
	int				size;
	Handle			text;
	char			*str = NULL;
	
	text = GetResource( 'TEXT', kHelpText );
	
	if ( !text ) FatalError( "Couldn't load help text." );
	
	size = GetHandleSize( text );
	SetHandleSize( text, size + 1 );	/* add one byte for trailing '\0'	*/
	HLock( text );
	
	str = (char *)*text;
	*(str + size) = '\0';				/* append '\0'						*/
	
	fputs( str, stdout );
	
	HUnlock( text );
	ReleaseResource( text );
}

#else

void DoHelp( void )
{
	int		c;
	FILE	*fp = fopen( "RedMech.help", "r" );
		
	if ( fp ) {
		while ( ( c = getc( fp ) ) != EOF ) putc( c, stdout );
		fclose( fp );
	} else {
		fputs( "# Couldn't load help text.\n", stderr );
		Usage();
	}
}

#endif
