/*
	Reduction.c

	RedMech is a tool for automatically reducing CnHm-mechanisms.
	Written by J. Goettgens, P. Terhoeven, R. Korff	
*/

#ifdef applec
#pragma load "RedMech.dump"
#else
#include "RedMechHeaders.h"
#endif


RedRecPtr NewReduction( void )
{
	static RedRecPtr	last = NULL;			/* Head of the list.			*/
	RedRecPtr			rr;
	
	rr = (RedRecPtr)malloc( sizeof(ReductionRecord) );
	if ( !rr ) FatalError( "Out of memory in NewReduction()." );
	memset( (char *)rr, 0, sizeof(*rr) );

	if ( !last )
		last = rr;
	else {
		last->next = rr;
		last = rr;
	}

	return rr;
}



RedRecPtr GetReductions( FILE *fp, int numOfReacs, SpeciesListPtr sl,
	ReactionPtr reacs, MatrixPtr m, int *numOfReducs )
{
	const char	prefix = 'w';
	char		*tmp = NULL;
	char		buffer[kBufSize+1], species[kMaxLen],
				label[kMaxLen], errorString[80];
	RedRecPtr	reduc = NULL, redList = NULL, item = NULL;
	int			line = 0;
	double		**a = m->mat;
	
	*numOfReducs = 0;
	while ( fgets( buffer, kBufSize, fp ) != NULL ) {		
#		ifdef applec
		RotateCursor( 32 );
#		endif

		++line;									/* increment the line count		*/
		if ( EmptyLine( buffer ) ) continue;	/* skip comment or empty lines	*/
		if ( sscanf( buffer, "%s%s", species, label ) != 2 ) 
			ParseError( line, "illegal input line" );
		UpperString( species );
		LowerString( label );
		
		/* Check for a leading 'w' in the reaction label.						*/
		tmp = strchr( label, prefix );
		if ( !tmp ) {
			sprintf( errorString, "illegal reaction label: \"%s\"!", label );
			ParseError( line, errorString );
		}
		++tmp;								/* Advance past the 'w'.			*/
		reduc = NewReduction();
		if ( !redList ) 
			redList = reduc;				/* Store head of list in redList.	*/
		++(*numOfReducs);					/* Count number of steady state.	*/

		strcpy( reduc->species, species );
		strcpy( reduc->label, tmp );
		reduc->row = Row( species, sl );
		if ( reduc->row < 0 ) {
			sprintf( errorString, "illegal species \"%s\" found", species );
			ParseError( line, errorString );
		}
		reduc->column = Column( tmp, reacs, numOfReacs );
		if ( reduc->column < 0 ) {
			sprintf( errorString, "illegal label \"%s\" found", tmp );
			ParseError( line, errorString );
		}
		
		/* Check if steady state species occurs in reaction to be eliminated.	*/
		if ( fabs( a[reduc->row][reduc->column] ) < kTiny ) {
			sprintf( errorString, "species %s not in reaction labelled %s!",
				reduc->species, reduc->label );
			ParseError( line, errorString );
		}
		
		/* Check for duplicate species 											*/
		item = redList;
		do {
			if ( item == reduc ) continue;		/* skip current species			*/
			if ( strcmp( item->species, species ) == 0 ) {
				sprintf( errorString, "duplicate Species \"%s\" found", species );
				ParseError( line, errorString );
			}
		} while ( item = item->next );
	}
	
	fclose( fp );
	
	return( redList );
}


void FreeReductions( RedRecPtr redList )
{
	RedRecPtr	next = NULL;
	
	while ( redList ) {
		next = redList->next;
		free( redList );
		redList = next;
	}
}


void PrintReductions( RedRecPtr redList, FILE *fp, char *label )
{
	RedRecPtr	rr = redList;
	
	if ( !fp ) fp = stdout;

	fprintf( fp, "Steady state species and reaction rates to eliminate\n" );
	if ( label ) fprintf( fp, "%s\n", label );
	
	while ( rr ) {
#		ifdef applec
		SpinCursor( 1 );
#		endif
		fprintf( fp, "\t%-4s\tw%s\n", rr->species, rr->label );
		rr = rr->next;
	}
	fprintf( fp, "\n" );
}


ConsEqPtr NewConsEqArray( int n ) 
{
	ConsEqPtr	cons = NULL;

	cons = (ConsEqPtr)malloc( n * sizeof(ConsEq) );
	if ( !cons ) FatalError( "Memory allcoation failed in NewConsEqArray()." );
	memset( (char *)cons, 0, n * sizeof(ConsEq) );

	return cons;
}


ConsEqPtr ConsEqFromReductions( RedRecPtr redList,
	SpeciesListPtr list, MatrixPtr m )
{
	int			i, numOfSpecies = list->num;
	double		**a = m->mat;
	RedRecPtr	first = NULL;
	ConsEqPtr	cons = NULL;
	
	cons = NewConsEqArray( numOfSpecies );
	
	for ( i = 0; i < numOfSpecies; ++i ) {
		first = redList;
		while ( first ) {
			if ( first->row == i ) {
				cons[i].isSteadyState = TRUE;
				break;
			}
			first = first->next;
		}
		cons[i].speciesName = list->species[i];
		cons[i].row = a[i];
	}
	
	return cons;
}


void FreeConsEq( ConsEqPtr cons )
{
	free( cons );
}


void PrintConsEq( ConsEqPtr cons, int numOfConsEq, int numOfReacs,
	ReactionPtr reacs, FILE *fp, char *label )
{
	int		i, j, first;
	int		printSteadyState;
	double	coeff;
	char	*bptr = gBuffer;
	
	if ( !fp ) fp = stdout;

	fputs( "Species balance equations", fp );
	if ( label ) fprintf( fp, " (%s)", label );
	fputs( ":\n", fp );
	
	/*	Loop twice over all species balance equations: 
			first all non-steady-state species
			second all steady-state species										*/
	for ( printSteadyState = 0; printSteadyState < 2; ++printSteadyState ) {
		for ( i = 0; i < numOfConsEq; ++i ) {
			first = TRUE;
			if ( printSteadyState ) {
				if ( !cons[i].isSteadyState ) continue;
			} else {
				if ( cons[i].isSteadyState ) continue;
			}
			
			/*bptr +=*/ sprintf( bptr, "\tL([%s]) = ", cons[i].speciesName );
			bptr += strlen(bptr);
			/*bptr +=*/ sprintf( bptr, "%s", ( cons[i].isSteadyState ) ? "0 = " : "" );
			bptr += strlen(bptr);

			for ( j = 0; j < numOfReacs; ++j ) {

				if ( fabs( coeff = cons[i].row[j] ) > kTiny ) {
					if ( bptr - gBuffer > kLineLength ) {
						fprintf( fp, "%s\n\t\t", bptr = gBuffer );
					}

					if ( coeff > 0.0 ) {
						if ( fabs(coeff - 1.0) < kTiny ) {
							/*bptr +=*/ sprintf( bptr, "%sw%s ",
								(first) ? "" : "+ ", reacs[j].label );
							bptr += strlen(bptr);
						} else {
							/*bptr +=*/ sprintf( bptr, "%s%g w%s ", 
								(first) ? "" : "+ ", coeff, reacs[j].label );
							bptr += strlen(bptr);
						}
						first = FALSE;
					} 
					else {
						if ( fabs(coeff + 1.0) < kTiny ) {
							/*bptr +=*/ sprintf( bptr, "- w%s ", reacs[j].label );
							bptr += strlen(bptr);
						} else {
							/*bptr +=*/ sprintf( bptr, "- %g w%s ", -coeff, reacs[j].label );
							bptr += strlen(bptr);
						}
						first = FALSE;
					}
#					ifdef applec
					SpinCursor( 1 );
#					endif
				}
			}
			fprintf( fp, "%s\n", bptr = gBuffer );
		}
		fputs( "\n", fp );
	}
}

