/*
	SpeciesList.c: Routines that deal with the list of species.

	RedMech is a tool for automatically reducing CnHm-mechanisms.
	Written by J. Goettgens, P. Terhoeven, R. Korff	
*/

#ifdef applec
#pragma load "RedMech.dump"
#else
#include "RedMechHeaders.h"
#endif


const char	*gThirdBody = "M";					/* Symbol for third body species.	*/
const char	gTheElems[] = { 'H', 'O', 'C', 'N', 'M' };

int	Row( char *speciesName, SpeciesListPtr list )
{
	int		i, n = list->num;
	
	UpperString( speciesName );					/* Normalize the species name.		*/
	
	for ( i = 0; i < n; ++i ) {					/* Sufficient for government work.	*/
		if ( strcmp( speciesName, list->species[i] ) == 0 ) return i;
	}

	return -1;									/* species name not found			*/
}


SpeciesListPtr AssembleSpeciesList( FILE *fp, int *numOfReacs )
{
	const char		*out_of_mem = "Out of memory in AssembleSpeciesList().";
	const int		kNameLen = 20;				/* Maximum length of species name.	*/
/*	const int		kNameLen = 8;				 Maximum length of species name.	*/
	SpeciesListPtr	sl = NULL;
	Token			theToken;
	char			*end = NULL;
	int				line;
	int				notDone = TRUE;
	
	sl = (SpeciesListPtr)malloc( sizeof(SpeciesList) );
	if ( !sl ) FatalError( out_of_mem );
	
	memset( (char *)sl, 0, sizeof(SpeciesList) );
	
	sl->names = (char *)malloc( kMaxSpecies * kNameLen );
	if ( !sl->names ) FatalError( out_of_mem );
	
	sl->cur = sl->names;						/* cur points to 1st char in names.	*/
	end = sl->cur + kMaxSpecies * kNameLen;		/* End of memory pool.				*/
		
	*numOfReacs = 0;							/* Reset the reaction count.		*/
	/*	Set line = 0 to reset Scanner's line count 			.						*/
	line = 0;
	
	/*	Now scan the file for identifiers.											*/
	do {
		char	*ident;
		int		i, len, isInList;

		Scanner( &theToken, fp, &line );
		
		switch ( theToken.type ) {				/* Count the number of reactions.	*/
			case kReactionDirection:
				++(*numOfReacs);
				break;
			case kIdentifier:					/* Handle identifiers only.			*/
				ident = theToken.string;
				isInList = FALSE;
				UpperString( ident );			/* Normalize the identifier.		*/
				for ( i = 0; i < sl->num; ++i ) {
					if ( strcmp( sl->species[i], ident ) == 0 ) {
						isInList = TRUE;		/* Found it.						*/
						break;
					}
				}
				if ( isInList ) break;			/* Skip this species.				*/
				if ( strcmp( ident, gThirdBody ) == 0 ) break;
				len = strlen( ident ) + 1;
				if ( sl->cur + len > end ) 
					FatalError( "Out of name space, now entering twilight zone." );
				sprintf( sl->cur, "%s", ident );
				sl->species[sl->num++] = sl->cur;
				sl->cur += len;
				break;
			case kDone:
				notDone = FALSE;
				break;
			default:
				/*	Do nothing.														*/
				break;
		}
	} while ( notDone );
	
	return sl;
}


void FreeSpeciesList( SpeciesListPtr p )
{
	if ( p->names ) free( p->names );
	free( p );
}


char *SpeciesName( SpeciesListPtr list, int index )
{
	if ( index < 0 || index > list->num-1 ) 
		FatalError( "Index out of range in SpeciesName()." );
		
	return list->species[index];
}


void PrintSpeciesList( SpeciesListPtr list, int numOfReacs,
						  FILE *fp, char *label )
{
	int		i, num = list->num;
	
	if ( !fp ) fp = stdout;
	fprintf( fp, "%d species found in %d reactions\n", num, numOfReacs );
	if ( label ) fprintf( fp, "%s\n", label );
	
	for ( i = 0; i < num; ++i ) {
#		ifdef applec
		SpinCursor( 1 );
#		endif
		fprintf( fp, "\t%s\n", list->species[i] );
	}
	fprintf( fp, "\n" );
}


int CountElements( SpeciesListPtr list )
{
#ifdef mips
	int			elemsFound[kElems];
#else
	int			elemsFound[kElems-1] = { 0, 0, 0, 0 };	/* skip third body			*/
#endif
	int			numOfElems = 0, i, j, num = list->num;
	const char	*pos = NULL;

#ifdef mips
	for ( i = 0; i < kElems; ++i ) elemsFound[i] = 0;
#endif
	
	for ( i = 0; i < num; ++i ) {

		pos = StripIsomerePrefix( list->species[i] );	/* Handle isomeres.			*/

		do {
			if ( isalpha(*pos) ) { 
				for ( j = 0; j < kElems-1; ++j ) {
					if ( *pos == gTheElems[j] ) {
						++elemsFound[j];
						break;
					}
				}
			}
		} while ( *(++pos) );
	}
	for ( i = 0; i < kElems-1; ++i ) numOfElems += ( elemsFound[i] > 0 );
	
	return numOfElems;
}
